/*------------------------------------------------------------------------------
 * Description: (Try to) Extract images from PPS files
 * Author     : Balazs Molnar, Hungary, mbdev@freemail.hu
 *------------------------------------------------------------------------------
 * $Id: ppsei.c,v 1.4 2011/07/03 21:14:10 mbdev Exp $
 *------------------------------------------------------------------------------
 *
 * Copyright (C) 2006, Balazs Molnar.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *------------------------------------------------------------------------------
 *
 * $Log: ppsei.c,v $
 * Revision 1.4  2011/07/03 21:14:10  mbdev
 * added gif parsing improvement: thx to chritophe duray
 *
 * Revision 1.3  2008/10/09 12:24:14  mbdev
 * Added GIF support + some bugfix (patch of Chris)
 *
 * Revision 1.2  2006/01/30 22:53:13  mbdev
 * added copyright
 *
 * Revision 1.1.1.1  2006/01/30 22:26:48  mbdev
 * added under CVS
 *
 *------------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>    /* basename() */
#include <string.h>    /* strncpy, ... */


/* how many image type is known: 1:JPG, 2:PNG , 3 GIF new format, 4 GIF old format*/
#define KNOWN_IMAGES 4

/* basename of the executable */
char exe[100];
const char * REVISION="REVISION:$Revision: 1.4 $" ;

void usage(void);

void
usage() {
  printf("%s: try to extract images from microsoft's PPS/PPT files\n\n", exe);

  printf("Usage: ppsei [-h|--help] <ppsfile-1> [... <ppsfile-n>]\n");
  printf("   -h       : this short help\n");
  printf("   <ppsfile>: name of the PPS/PPT file to process\n\n");
  printf("Currently only JPEG,PNG and GIF files can be extracted. Because the pps\n");
  printf("fileformat is not so well-defined as the gnumeric :-), there is no\n");
  printf("guarantee, that all images will be grabbed correctly. But the experience is,\n");
  printf("that it works.\n");
  printf("The extracted files will be placed under the current working directory. With\n");
  printf("the filename <ppsfile>_<index>.(jpg|png|gif), where ppsfile is the filename\n");
  printf("of the processed pps (with the extension), and index is a counter\n");
  printf("restarted at every input file (f.e. X.pps, ppsei -> X.pps_01.jpg)\n");
}

int
main(int argc, char * argv[]) {
  FILE   *pps, *image;
  char   *buffer;
  struct stat filestat;
  char   imagename[256];
  int    i, j, d, imagebegin = 0, imageend = 0, ppsi = 1, I, count;
  long int    filesize;
  int    t, typeindex = 0;

  /*
   * use 'hexedit some.pps' to see this bytes :-)
   * match arrays, this byte sequences will be searched in the buffer
   * 0xAA: "END of array"
   */
  char JPGH[] = { 0xFF, 0xD8, 0xFF, 0xE0,
                  0x00, 0x10, 0x4A, 0x46, 0x49, 0x46, 0xAA };
  char JPGE[] = { 0xFF, 0xD9, 0xAA };
  char PNGH[] = { 0x89, 0x50, 0x4E, 0x47,
                  0x0D, 0x0A, 0x1A, 0x0A,
                  0x00, 0x00, 0x00, 0xAA };
  char PNGE[] = { 0x00, 0x00, 0x00, 0x00,
                  0x49, 0x45, 0x4E, 0x44,
                  0xAE, 0x42, 0x60, 0x82, 0xAA };

  char GIFH[] = { 0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0xAA };
  char GIFE[] = { 0x00, 0x3B, 0xAA };

  char GIOH[] = { 0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0xAA };
  char GIOE[] = { 0x00, 0x3B, 0xAA };


  /* extensions by creating the appropriate image filename */
  char* TYPES[]  = { NULL, "jpg", "png", "gif", "gif"};

  /* array-heads of known image type's begin/end match-arrays */
  char* HEADER[] = { NULL, JPGH, PNGH, GIFH, GIOH};
  char* END[]    = { NULL, JPGE, PNGE, GIFE, GIOE};

  /* exe: basename of this program */
  strncpy(exe, basename(argv[0]),100);

  if (argc == 1) {
    printf("Error: argument expected: see usage: with -h or --help\n");
    exit(1);
  }

  if ( (! strcmp(argv[1], "-h")) ||
       (! strcmp(argv[1], "-help")) || (! strcmp(argv[1], "--help"))) {
    usage();
    exit(0);
  }

  /* loop over all arguments (they are probably pps files */
  while (ppsi < argc) {

    if (stat(argv[ppsi],&filestat) ) {
      perror("stat of %s\n");
      continue;
    }
    filesize=filestat.st_size;

    if (! (pps = fopen(argv[ppsi],"r"))) {
      printf("ERROR: cannot open file: %s!\n", argv[ppsi]);
      continue;
    }

    printf("Allocating %ld bytes...\n", filesize);
    buffer=malloc(filesize+1);
    if (buffer == NULL) {
      printf("ERROR: cannot allocate memory (%ld bytes)\n",filesize);
      continue;
    }

    /* I: the numer of found images (will be reseted to 0 by every new PPS) */
    I = 0;

    while (!feof(pps)) {
      while ((!feof(pps)) && (count = fread(buffer, 1, filesize, pps))) {
        /* here the pps was read into buffer (so only 1 loop will be mase)
         * i: position in the PPS byte-stream
         */
        i = 0;
        /* let's start searching for matches */
        while (i < count) {
          /* check IMAGE end
           * typeindex hold the founded image type, this can be:
           * 0: not found, 1: jpeg, 2: png
           */
          if (typeindex) {
            d = 0; j = 0 ;
            /* d: position in the END array: at where is the match currently
             * j: position in the ppsfile  : at where is the match currently
             * (i + d) < count)     -> end of file
             * (END[typeindex][j]   -> end of END array
             */
            while (((i + d) < count) &&
                   (END[typeindex][j] != (char) 0xAA) &&
                   (END[typeindex][j] == buffer[i + d])) {
              d++; j++;
            }

            if (END[typeindex][j] == (char) 0xAA) { /* image end found */

              /* new read position in the buffer (skip the end pattern) */
              i += d - 1;

              /* this is the endpos of the founded image */
              imageend = i ;

              /* image found: new index for the fielname */
              I++;

              sprintf(imagename, "%s_%02d.%s", argv[ppsi], I, TYPES[typeindex]);
              if (! (image = fopen(imagename, "w+"))) {
                printf("ERROR: cannot open image file: %s!\n", imagename);
                continue;
              }

              printf("%s image %d written at 0x%x bytes: %d\n",
                      TYPES[typeindex], I, imagebegin, imageend - imagebegin);
              fwrite(&buffer[imagebegin], 1, imageend - imagebegin + 1, image);
              fclose(image);
              /* lets start to find a new image */
              typeindex = 0;
            }
          }

          /* header check: t: typeindex, 0:noimage, 1:jpg, 2:png */
          t = 1 ;
          while (t < KNOWN_IMAGES + 1) {
            d = 0; j = 0;
            /*
             * d: position in the HEADER array: at where is the match currently
             * j: position in the ppsfile  : at where is the match currently
             * (i + d) < count)     -> end of file
             * (HEADER[t][j]       -> end of HEADER array
             */

            while (((i + d) < count) &&
                   (HEADER[t][j] != (char) 0xAA) &&
                   (HEADER[t][j] == buffer[i + d])) {
              d++; j++;
            }

            if (HEADER[t][j] == (char) 0xAA) { /* image begin found */
              /* image found */
              typeindex = t;

              /* this is the begin pos in the buffer  of the founded image: */
              imagebegin = i;

              /* new read position in the buffer (skip the end pattern): */
              i += d - 1 ;
              break;
            }
            t++;
          }

          i++;

        }/* (i < count) */
      }/* (!feof) */
    }

    free(buffer);

    if (I)
      printf("%s: processed: '%s', extracted %d image(s).\n", exe, argv[ppsi], I);
    else
      printf("%s: processed: '%s', no image found.\n", exe, argv[ppsi]);

    /* process next inputfile */
    ppsi++;
  }

  exit(0);
}
